/*jslint node: true */
/*jshint esversion: 6 */
"use strict";

let express = require("express");
let paginate = require("express-paginate");
let Route = require("route-parser");
let debug = require("debug");
let {
    unless
} = require("express-unless");
let vhost = require("vhost");
let define = require("define-js");
let parse = require("parseparams");
let {
    URL,
    URLSearchParams
} = require("url");
let crypto = require("crypto");
let {
    BusinessBehaviourType,
    BusinessBehaviour
} = require("behaviours-js");
let {
    OFFLINESYNC,
    OFFLINEACTION,
    ONLINESYNC,
    ONLINEACTION
} = BusinessBehaviourType;
let {
    businessController
} = require("./controller.js");
let {
    getLogBehaviour
} = require("./log.js");
let {
    scheduleBehaviour
} = require("./schedule.js");
let {
    getInputObjects,
    setResponse,
    respond,
    getSignature,
    setSignature,
    getRequest
} = require("./utils.js");

debug.enable("backend:*");
debug = debug("backend:behaviour");

let backend = module.exports;

let app = backend.app = express();

backend.serve = express.static;

let join = backend.join = function (s1, s2) {

    let fromIndex = s2.startsWith("/") ? 1 : 0;
    let toIndex = s1.length;
    if (s1.endsWith("/")) toIndex--;
    s1 = s1.substr(0, toIndex) + "/";
    s2 = s2.substr(fromIndex);
    let url = new URL(...[
        s2,
        new URL(s1, "resolve://")
    ]);
    if (url.protocol === "resolve:") {

        let { pathname, search, hash } = url;
        return pathname + search + hash;
    }
    return url.toString();
};

let compare = backend.compare = function () {

    let [route1, route2] = arguments;
    let route = route2;
    let varying = !!(route1 && route1.path);
    if (varying) {

        varying = route1.path.indexOf(":") > -1;
        varying |= route1.path.indexOf("*") > -1;
    }
    if (varying) route = route1;
    if (route === route2) {

        route2 = route1;
        route1 = route;
    }
    if (route && route.path) {

        route = new Route(route.path);
    }
    let path1 = route1 && route1.path;
    let path2 = route2 && route2.path;
    let method1 = (route1 && route1.method) || "";
    method1 = method1.toLowerCase();
    let method2 = (route2 && route2.method) || "";
    method2 = method2.toLowerCase();
    let matched = !!route;
    if (matched) {

        matched &= !!route.match(path2 || " ");
    }
    matched |= path1 === path2;
    matched &= method1 === method2;
    return matched;
};

let resolve = backend.resolve = function () {

    let [prefix, suffix, path] = arguments;
    let prefixed = typeof prefix === "string";
    if (prefixed) {

        prefixed &= path.startsWith(prefix);
    }
    if (prefixed && typeof suffix === "string") {

        return join(prefix, suffix);
    } else return suffix || prefix;
};

let defaultPrefix = "/";

let types = {

    database: OFFLINESYNC,
    database_with_action: OFFLINEACTION,
    integration: ONLINESYNC,
    integration_with_action: ONLINEACTION
};

let routers = {};

let events = {};

let emitters = {};

let behaviours = {

    behaviours: {

        method: "GET",
        path: "/behaviours"
    }
};

let BEHAVIOURS = {};

let defaultOperations = {};

let defaultRemotes = {};

let FetchBehaviours = {};

let LogBehaviours = {};

let upgradePlugins = {};

backend.behaviour = function (path, config) {

    let no_app = typeof app !== "function";
    if (!no_app) {

        no_app |= typeof app.use !== "function";
    }
    if (no_app) throw new Error("Invalid express app");
    if (typeof path === "object") {

        config = path;
        path = config.path;
    }
    if (typeof config !== "object" || !config) {

        config = {};
    }
    let no_operations = !config.operations;
    if (!no_operations) {

        let { operations } = config;
        no_operations |= typeof operations !== "object";
    }
    if (no_operations) config.operations = {};
    return function (options, getConstructor) {

        if (typeof options !== "object") {

            throw new Error("Invalid definition object");
        }
        if (typeof options.inherits === "function") {

            let { prototype } = options.inherits;
            if (!(prototype instanceof BusinessBehaviour)) {

                throw new Error("Super behaviour should " +
                    "inherit from BusinessBehaviour");
            }
            options = Object.assign(Object.keys(...[
                BEHAVIOURS
            ]).reduce(function (ȯptions, name) {

                let { constructor } = BEHAVIOURS[name];
                if (constructor == options.inherits) {

                    return BEHAVIOURS[name].options;
                }
                return ȯptions;
            }, {}), options);
        }
        no_operations = !options.operations;
        if (!no_operations) {

            let { operations } = options;
            no_operations |= typeof operations !== "object";
        }
        if (no_operations) options.operations = {};
        options.operations = Object.assign(...[
            {},
            defaultOperations,
            config.operations,
            options.operations
        ]);
        let no_type = typeof options.type !== "string";
        if (!no_type) {

            no_type |= types[options.type] === undefined;
        }
        if (no_type) options.type = "database";
        let no_version = typeof options.version !== "string";
        if (!no_version) {

            no_version |= options.version.length === 0;
        }
        if (no_version) {

            throw new Error("Invalid behaviour version");
        }
        if (typeof getConstructor !== "function") {

            throw new Error("Invalid constructor");
        }
        if (!Array.isArray(options.events)) {

            if (options.events) {

                debug('Events should be array or use event');
            }
            options.events = [];
        }
        if (typeof options.event === "function") {

            options.events.push(options.event);
        }
        options.events = options.events.filter(...[
            function (event) {

                let valid = typeof event === "function";
                if (!valid) {

                    valid = typeof event === "string";
                    if (valid) {

                        valid &= event.length > 0;
                    }
                }
                return valid;
            }
        ]);
        let named = typeof options.name === "string";
        if (named) {

            named &= options.name.length > 0;
        }
        let uniquelyNamed = function () {

            let { skipSameRoutes } = config;
            if (named && BEHAVIOURS[
                options.name
            ] && skipSameRoutes !== true) {

                throw new Error("Duplicate behavior name: " +
                    options.name + ". Make sure names are " +
                    "unique and not numerical");
            }
            return named && !BEHAVIOURS[options.name];
        }();
        let BehaviourConstructor = define(...[
            getConstructor
        ]).extend(getLogBehaviour(...[
            options,
            config,
            types,
            BEHAVIOURS,
            defaultRemotes,
            FetchBehaviours,
            LogBehaviours,
            function (room) {

                return emitters[room];
            }
        ])).defaults({

            type: types[options.type]
        });
        if (options.fetcher) {

            let fetcher = "";
            if (typeof options.fetcher === "string") {

                fetcher = options.fetcher;
            }
            FetchBehaviours[fetcher] = BehaviourConstructor;
        }
        if (options.logger) {

            let logger = "";
            if (typeof options.logger === "string") {

                options.logger = logger;
            }
            LogBehaviours[logger] = BehaviourConstructor;
        }
        scheduleBehaviour(...[
            options,
            BehaviourConstructor,
            types,
            FetchBehaviours
        ]);
        if (uniquelyNamed) {

            if (options.name === "behaviours") {

                throw new Error("behaviours is a reserved name");
            }
            let middleware = typeof options.path === "string";
            if (middleware) {

                middleware &= options.path.length > 0;
            }
            let routing = middleware;
            routing &= typeof options.method === "string";
            if (routing) {

                let method = options.method.toLowerCase();
                routing &= typeof app[method] === "function";
            }
            let polling = routing && Object.keys(...[
                types
            ]).indexOf(options.type) > 1;
            if (!Array.isArray(options.plugins)) {

                options.plugins = [];
            }
            if (typeof options.plugin === "function") {

                options.plugins.push(options.plugin);
            }
            let request_plugins = options.plugins.filter(...[
                function (plugin) {

                    let valid = typeof plugin === "function";
                    if (valid) {

                        valid &= parse(plugin)[0] !== "out";
                    }
                    return valid;
                }
            ]);
            let upgradePlugin;
            let upgrading = request_plugins.length > 0;
            if (upgrading) {

                upgradePlugin = request_plugins.find(...[
                    function (plugin) {

                        let [last] = parse(plugin).reverse();
                        return last === "head";
                    }
                ]);
                upgrading &= !!upgradePlugin;
            }
            if (upgrading) {

                upgradePlugins[options.name] = upgradePlugin;
            }
            let response_plugin = options.plugins.reduce(...[
                function (response_plugin, plugin) {

                    let valid = typeof plugin === "function";
                    if (valid) {

                        valid &= parse(plugin)[0] === "out";
                    }
                    if (valid) return plugin;
                    return response_plugin;
                },
                undefined
            ]);
            let prefix;
            let pathing = typeof path === "string";
            if (pathing) {

                pathing &= path.length > 0;
            }
            if (pathing) {

                if (config.overwritePath) prefix = path;
                else prefix = join(defaultPrefix, path);
            } else {

                let no_overwrite = !config.overwritePath;
                no_overwrite &= defaultPrefix !== "/";
                if (no_overwrite) {

                    prefix = defaultPrefix;
                }
            }
            if (options.events.length > 0 && join(...[
                prefix,
                options.path
            ]) == join(prefix, "/events")) {

                throw new Error("Invalid path. " +
                    join(prefix, options.path) +
                    " is reserved route");
            }
            BEHAVIOURS[options.name] = {

                options: Object.assign({ prefix }, options),
                constructor: BehaviourConstructor
            };
            let behaviour_runner = function () {

                let [
                    req,
                    res,
                    next,
                    inputObjects,
                    er
                ] = arguments;
                let onClose;
                let signature = getSignature(req);
                let response = {

                    behaviour: options.name,
                    version: options.version
                };
                if (polling) {

                    let time = new Date(signature).getTime();
                    response.signature = time;
                    setSignature(req, res, next, response);
                    if (typeof signature === "number") return;
                }
                if (options.paginate) {

                    inputObjects.paginate = true;
                    inputObjects.page = req.query.page;
                    inputObjects.limit = req.query.limit;
                }
                let behaviour = new BehaviourConstructor({

                    name: options.name,
                    type: types[options.type],
                    priority: options.priority || 0,
                    timeout: options.timeout,
                    inputObjects
                }, function (name, room) {

                    if (!req.session) return;
                    let event = events[name];
                    if (event && event[room]) {

                        let { id } = req.session;
                        let client = event[room][id];
                        if (client) return client.id;
                    }
                }, function () {

                    return req.complete;
                });
                let behaviour_callback = function () {

                    let [
                        result,
                        error
                    ] = arguments;
                    let request = getRequest(...[
                        req,
                        res,
                        next,
                        response
                    ]);
                    if (!request) {

                        if (polling) setResponse(...[
                            behaviour_callback.bind(...[
                                null,
                                result,
                                error
                            ]),
                            response
                        ]);
                        return;
                    }
                    if (polling) delete response.signature;
                    let failing = typeof error === "object";
                    failing |= typeof result !== "object";
                    if (failing) {

                        if (error && !error.behaviour) {

                            error.behaviour = options.name;
                        }
                        if (error && !error.version) {

                            error.version = options.version;
                        }
                        request.next(...[
                            error || er || new Error("Error " +
                                "while executing " +
                                options.name +
                                " behaviour, version " +
                                options.version + "!")
                        ]);
                    } else {

                        let responding = !response_plugin;
                        if (!responding) {

                            responding |= !response_plugin(...[
                                result,
                                request.req,
                                request.res,
                                request.next
                            ]);
                        }
                        if (responding) {

                            response.response = result;
                            if (options.paginate) {

                                let {
                                    modelObjects: page
                                } = result;
                                if (page) {

                                    response.response = page;
                                }
                            }
                            let { length } = options.events;
                            let eventful = length > 0;
                            eventful &= !!req.session;
                            if (eventful) {

                                let _ = crypto.randomBytes(48);
                                let token = _.toString("base64");
                                response.events_token = token;
                                ({ events: _ } = options);
                                response.events = _.map(...[
                                    function (event) {

                                        let room = event;
                                        _ = typeof event;
                                        if (_ === "function") {

                                            room = event(...[
                                                options.name,
                                                inputObjects
                                            ]);
                                        }
                                        let jsonify = !!room;
                                        _ = typeof room;
                                        jsonify &= _ === "object";
                                        if (jsonify) {

                                            let {
                                                stringify
                                            } = JSON;
                                            return stringify(room);
                                        }
                                        return room;
                                    }
                                ]).filter(function (room) {

                                    _ = typeof room;
                                    let valid = _ === "string";
                                    if (valid) {

                                        valid &= !!room.trim();
                                    }
                                    if (valid) {

                                        let event = events[
                                            options.name
                                        ];
                                        if (!event) {

                                            event = events[
                                                options.name
                                            ] = {};
                                        }
                                        if (!event[room]) {

                                            event[room] = {};
                                        }
                                        let { id } = req.session;
                                        event[room][id] = {

                                            token,
                                            date: new Date(),
                                            count: 0
                                        };
                                        return true;
                                    }
                                    return false;
                                });
                            }
                            if (options.paginate) {

                                let {
                                    pageCount: page
                                } = result;
                                if (typeof page !== "number") {

                                    page = 1;
                                }
                                let _ = paginate.hasNextPages(...[
                                    request.req
                                ])(page);
                                response.has_more = _;
                            }
                            let { returns } = options;
                            if (typeof returns !== "function") {

                                if (!setResponse(...[
                                    returns,
                                    !routing,
                                    request,
                                    response
                                ])) request.next();
                            } else returns(...[
                                request.req,
                                request.res,
                                result,
                                error,
                                function (outputObjects) {

                                    respond(...[
                                        request.res,
                                        outputObjects
                                    ]);
                                }
                            ]);
                        }
                    }
                    if (onClose) {

                        req.socket.removeListener(...[
                            "close", onClose
                        ]);
                    }
                };
                let fetching = "";
                if (typeof options.fetching === "string") {

                    fetching = options.fetching;
                }
                let FetchBehaviour = FetchBehaviours[fetching];
                if (options.fetcher) {

                    FetchBehaviour = BehaviourConstructor;
                }
                let { queue } = options;
                if (typeof queue === "function") {

                    queue = queue(options.name, inputObjects);
                }
                let cancel = businessController(...[
                    options.name,
                    queue,
                    options.database,
                    options.storage,
                    options.fetcher || options.fetching,
                    FetchBehaviour,
                    options.memory,
                    options.operations
                ]).runBehaviour(...[
                    behaviour,
                    options.paginate ? function () {

                        let [
                            property,
                            superProperty
                        ] = arguments;
                        let page = {

                            modelObjects: "modelObjects",
                            pageCount: "pageCount"
                        };
                        let map = { options };
                        if (typeof map === "function") {

                            let mapped = map(...[
                                property,
                                superProperty
                            ]);
                            if (mapped) return mapped;
                        }
                        return page[property];
                    } : options.map,
                    behaviour_callback
                ]);
                req.socket.on("close", onClose = function () {

                    let _ = typeof cancel;
                    let cancelling = _ === "function";
                    cancelling &= !polling;
                    if (cancelling) {

                        cancelling &= !req.readableEnded;
                        if (!cancelling) {

                            cancelling |= !res.writableEnded;
                        }
                    }
                    if (cancelling) {

                        cancel();
                        debug("Request aborted and " +
                            "behaviour cancelled");
                    }
                });
            };
            let request_handler = function (req, res, next) {

                if (typeof options.parameters !== "function") {

                    if (!routing || req.complete) {

                        getInputObjects(...[
                            options.parameters,
                            Object.keys(behaviours).map(...[
                                function (name) {

                                    let {
                                        [name]: ȯptions
                                    } = behaviours, suffix;
                                    if (ȯptions) {

                                        suffix = ȯptions.path;
                                    }
                                    return resolve(...[
                                        prefix,
                                        suffix,
                                        req.path
                                    ]);
                                }]
                            ),
                            req,
                            function (inputObjects) {

                                behaviour_runner(...[
                                    req,
                                    res,
                                    next,
                                    inputObjects
                                ]);
                            }
                        ]);
                    } else req.socket.on(...[
                        "end",
                        request_handler.bind(null, req, res, next)
                    ]);
                } else options.parameters(...[
                    req,
                    res,
                    function (inputObjects, er) {

                        if (req.complete) behaviour_runner(...[
                            req,
                            res,
                            next,
                            inputObjects,
                            er
                        ]); else throw new Error("Parameters" +
                            " callback function called before" +
                            " all request data consumed");
                    }
                ]);
            };
            let filtering = Array.isArray(options.unless);
            filtering |= Array.isArray(options.for);
            if (filtering) {

                request_handler.unless = unless;
                request_handler = request_handler.unless({

                    custom(req) {

                        let exceptions = [];
                        if (Array.isArray(options.for)) {

                            exceptions = options.for;
                        } else options.for = undefined;
                        if (Array.isArray(options.unless)) {

                            exceptions = options.unless;
                        } else options.unless = undefined;
                        exceptions = exceptions.filter(...[
                            function (name) {

                                let {
                                    [name]: ȯptions
                                } = behaviours, suffix, method;
                                if (ȯptions) {

                                    suffix = ȯptions.path;
                                    method = ȯptions.method;
                                }
                                return compare({

                                    path: resolve(...[
                                        prefix,
                                        suffix,
                                        req.path
                                    ]),
                                    method
                                }, {

                                    path: req.path,
                                    method: req.method
                                });
                            }
                        ]).length;
                        if (options.unless) {

                            return exceptions > 0;
                        }
                        return exceptions === 0;
                    }
                });
            }
            filtering = typeof options.host === "string";
            if (filtering) {

                filtering &= options.host.length > 0;
            }
            if (filtering) {

                request_handler = vhost(...[
                    options.host,
                    request_handler
                ]);
                let plugins = request_plugins;
                if (plugins.length > 0) {

                    request_plugins = plugins.map(...[
                        function (plugin) {

                            return vhost(...[
                                options.host, plugin
                            ]);
                        }
                    ]);
                }
            } else if (request_plugins.length > 0) {

                let plugins = request_plugins;
                request_plugins = plugins.map(...[
                    function (plugin) {

                        return function (req, res, next) {

                            plugin(req, res, next);
                        };
                    }
                ]);
            }
            if (routing) {

                let names = Object.keys(behaviours);
                let {
                    skipSameRoutes
                } = config;
                if (!skipSameRoutes && names.some(...[
                    function (name) {

                        let {
                            [name]: ȯptions
                        } = behaviours;
                        return compare({

                            path: ȯptions.path,
                            method: ȯptions.method
                        }, {

                            path: options.path,
                            method: options.method
                        });
                    }
                ])) {

                    throw new Error("Duplicated behavior" +
                        " path: " + options.path);
                }
                let router = app;
                let prefixing = typeof prefix === "string";
                if (prefixing) {

                    prefixing &= prefix.length > 0;
                }
                if (prefixing) {

                    router = routers[prefix];
                    if (!router) {

                        router = express.Router({

                            caseSensitive: true,
                            mergeParams: true,
                            strict: true
                        });
                        router.use(paginate.middleware(10, 50));
                        app.use(prefix, router);
                        routers[prefix] = router;
                    }
                }
                router = router[
                    options.method.toLowerCase()
                ].bind(router);
                if (request_plugins.length > 0) router(...[
                    options.path,
                    ...request_plugins,
                    request_handler
                ]); else router(options.path, request_handler);
                behaviours[options.name] = {

                    version: options.version,
                    method: options.method,
                    path: options.path,
                    host: options.host,
                    events: options.events.length > 0,
                    prefix,
                    origins: options.origins,
                    maxAge: options.maxAge,
                    parameters: function () {

                        let { parameters } = options;
                        if (typeof parameters !== "function") {

                            return parameters;
                        }
                    }(),
                    returns: function () {

                        let { returns } = options;
                        if (typeof returns !== "function") {

                            return returns;
                        }
                    }()
                };
            } else if (middleware) {

                let route = options.path;
                let prefixing = typeof prefix === "string";
                if (prefixing) {

                    prefixing &= prefix.length > 0;
                }
                if (prefixing) {

                    route = join(prefix, options.path);
                }
                if (request_plugins.length > 0) app.use(...[
                    route,
                    ...request_plugins,
                    request_handler
                ]); else app.use(route, request_handler);
            } else {

                if (request_plugins.length > 0) app.use(...[
                    ...request_plugins,
                    request_handler
                ]); else app.use(request_handler);
            }
        } else BEHAVIOURS[Object.keys(BEHAVIOURS).length + 1] = {

            options,
            constructor: BehaviourConstructor
        };
        return BehaviourConstructor;
    };
};

backend.BehavioursServer = function () {

    let [prefix, parser, remotes, operations] = arguments;
    if (operations && typeof operations === "object") {

        defaultOperations = operations;
    }
    if (remotes && typeof remotes === "object") {

        defaultRemotes = remotes;
    }
    let default_prefixing = defaultPrefix === "/";
    default_prefixing &= typeof prefix === "string";
    if (default_prefixing) {

        default_prefixing &= prefix.length > 0;
    }
    if (default_prefixing) defaultPrefix = prefix;
    if (!prefix) prefix = defaultPrefix;
    let prefixing = typeof prefix === "string";
    if (prefixing) {

        prefixing &= prefix.length > 0;
    }
    app.get(function () {

        if (prefixing) return join(prefix, "/behaviours");
        return "/behaviours";
    }(), function (_, res) {

        respond(res, behaviours, parser);
    });
    let validate_path = function (behaviour, path) {

        let behaviour_prefix = behaviour.prefix;
        if (!behaviour_prefix) {

            behaviour_prefix = defaultPrefix;
        }
        let behaviour_path = behaviour.path || "/";
        let routing = typeof behaviour.method === "string";
        if (routing) {

            let method = behaviour.method.toLowerCase();
            routing &= typeof app[method] === "function";
        }
        if (!routing) {

            behaviour_path = join(behaviour_path, "/*path");
        }
        return compare({

            path: resolve(...[
                behaviour_prefix,
                behaviour_path,
                path
            ])
        }, {

            path
        });
    };
    let validate_host = function (host, req, res) {

        let same_host = true;
        let filtering = typeof host === "string";
        if (filtering) {

            filtering &= host.length > 0;
        }
        if (filtering) vhost(host, function () { })(...[
            req,
            res,
            function () {

                same_host = false;
            }
        ]);
        return same_host;
    };
    this.upgrade = function (req, socket, head) {

        let names = Object.keys(BEHAVIOURS);
        let [
            path,
            query
        ] = (req.originalUrl || req.url).split("?");
        if (query) {

            query = new URLSearchParams(...[
                query
            ]).toString();
            if (names.indexOf(query.behaviour) > -1) {

                names = [query.behaviour];
            }
        }
        for (let i = 0; i < names.length; i++) {

            if (!upgradePlugins[names[i]]) continue;
            let behaviour = BEHAVIOURS[names[i]].options;
            let upgrading = validate_host(...[
                behaviour.host,
                req,
                socket
            ]);
            if (upgrading) {

                upgrading = validate_path(behaviour, path);
            }
            if (upgrading) {

                upgradePlugins[names[i]](...[
                    req,
                    socket,
                    undefined,
                    head
                ]);
                return true;
            }
        }
        return false;
    };
    this.validate = function (path, query) {

        let name = query.behaviour;
        let named = typeof name === "string";
        if (named) {

            named &= name.length > 0;
        }
        if (named) {

            let behaviour;
            if (BEHAVIOURS[name]) {

                behaviour = BEHAVIOURS[name].options;
            }
            let eventful = !!behaviour;
            if (eventful) {

                eventful &= !!behaviour.events;
            }
            if (eventful && compare({

                path: resolve(...[
                    behaviour.prefix,
                    "/events",
                    path
                ])
            }, {

                path
            })) return;
        }
        return new Error("Not found");
    };
    this.connect = function (socket) {

        let client;
        let name = socket.handshake.auth.behaviour;
        if (!name) {

            name = socket.handshake.query.behaviour;
        }
        let token = socket.handshake.auth.token;
        if (!token) {

            token = socket.handshake.query.token;
        }
        let id = (socket.handshake.session || {}).id;
        let authenticating = typeof name === "string";
        if (authenticating) {

            authenticating &= name.length > 0;
        }
        authenticating &= typeof token === "string";
        if (authenticating) {

            authenticating &= token.length > 0;
        }
        if (authenticating) {

            let behaviour;
            if (BEHAVIOURS[name]) {

                behaviour = BEHAVIOURS[name].options;
            }
            let eventful = !!behaviour;
            if (eventful) {

                eventful &= !!behaviour.events;
            }
            if (eventful && validate_host(...[
                behaviour.host,
                socket.request,
                socket
            ])) {

                let joined = false;
                let event = events[name];
                if (event) socket.on(...[
                    "join " + name,
                    function (room) {

                        if (event[room]) {

                            client = event[room][id];
                        }
                        if (client) {

                            if (client.id !== socket.id) {

                                client.count++;
                            }
                            let { date: dt } = client;
                            dt = dt.getTime();
                            dt = new Date().getTime() - dt;
                            let authenticated = dt < 60000;
                            if (client.token !== token) {

                                authenticated = false;
                            }
                            if (client.count !== 1) {

                                authenticated = false;
                            };
                            if (authenticated) {

                                let room_events = emitters[
                                    room
                                ];
                                if (!room_events) {

                                    room_events = emitters[
                                        room
                                    ] = {};
                                }
                                let ëmitters = room_events[
                                    name
                                ];
                                if (!ëmitters) {

                                    ëmitters = room_events[
                                        name
                                    ] = [];
                                }
                                if (!ëmitters.find(...[
                                    function () {

                                        let [{
                                            name: e_id
                                        }] = arguments;
                                        let {
                                            name: nsp_id
                                        } = socket.nsp;
                                        return e_id == nsp_id;
                                    }
                                ])) {

                                    ëmitters.push(socket.nsp);
                                }
                                if (client.id !== socket.id) {

                                    client.id = socket.id;
                                    socket.join(room);
                                    joined = true;
                                }
                            }
                        }
                    }
                ]);
                setTimeout(function () {

                    if (!joined) socket.disconnect(true);
                }, 60000);
                socket.once("disconnect", function () {

                    if (client) {

                        client.count--;
                        if (client.count === 0) {

                            client.date = new Date();
                        }
                    }
                });
                return;
            }
        }
        socket.disconnect(true);
    };
};

backend.routes = behaviours;
